<?php

/**
 * @file
 * Implements help hooks for the module.
 */

use Drupal\Core\Routing\RouteMatchInterface;

use Drupal\foldershare\Utilities\LinkUtilities;

use Drupal\foldershare_rest\Branding;

/**
 * Implements hook_help().
 *
 * This function responds to the help route for the module.
 */
function foldershare_rest_help(
  string $routeName,
  RouteMatchInterface $routeMatch) {
  //
  // The function returns a render array containing multiple help sections,
  // partly mandated by Drupal best practices.
  //
  // Intentionally use the same class names as the FolderShare module
  // so that the page gets the same styling.
  //
  // Check the route.
  $helpRoute = 'help.page.foldershare_rest';
  if ($routeName !== $helpRoute) {
    // The route is not for our help page. Do nothing.
    return;
  }

  // Check what is installed.
  // - REST must be installed since it is a dependency of this module.
  // - REST UI is optional.
  // - Help is optional.
  $mh = \Drupal::service('module_handler');
  $restUiInstalled = $mh->moduleExists('restui');
  $helpInstalled = $mh->moduleExists('help');

  if ($helpInstalled === TRUE) {
    $restHelpLink = LinkUtilities::createHelpLink(
      'rest',
      $mh->getName('rest'));
  }
  else {
    $restHelpLink = $mh->getName('rest');
  }

  if ($restUiInstalled === TRUE) {
    $restUiName = $mh->getName('restui');
  }
  else {
    $restUiName = 'REST UI';
  }

  if ($restUiInstalled === TRUE) {
    $restUiLink = LinkUtilities::createRouteLink(
      'restui.list',
      '',
      $restUiName);
  }
  else {
    $restUiLink = LinkUtilities::createProjectLink(
      'restui',
      $restUiName);
  }

  $foldershareName = $mh->getName('foldershare');
  $foldershareSettingsLink = LinkUtilities::createRouteLink(
    'entity.foldershare.settings');
  $foldersharePermissionsLink = LinkUtilities::createRouteLink(
    'user.admin_permissions',
    'module-foldershare',
    $foldershareName);

  //
  // About
  // -----
  // This section is recommended by Drupal best practices.
  $help = '';
  $help .= '<h4>' . t('About') . '</h4>';
  $help .= '<p>' . t(
    'This module provides REST web services for the %foldershareSettingsLink module. This enables users to use non-browser client applications to upload, download, create, delete, and modify shared files and folders managed by %foldershareName.',
    [
      '%foldershareSettingsLink' => $foldershareSettingsLink,
      '%foldershareName'         => $foldershareName,
    ]) . '</p>';

  //
  // Uses
  // ----
  // This section is recommended by Drupal best practices.
  $help .= '<h4>' . t('Uses') . '</h4>';
  $help .= '<p>' . t(
    'While a web browser provides interactive access to a web site, REST web services enable automatic scripted access as well as alternate client applications that may present customized user interfaces. For instance, the included command-line REST client provides a shell-like interface and Linux/BSD/macOS-style commands to access shared files (e.g. "ls", "cp", "rm"). This command-line client can be used interactively, or embedded in shell scripts that compute or process data and automatically post results to a web site.',
    [
      '%foldershareName' => $foldershareName,
    ]) . '</p>';

  //
  // Security
  // --------
  // Explain what is not supported.
  $help .= '<h4>' . t('Security') . '</h4>';
  $help .= '<p>' . t(
    "REST services only provide access to files and folders managed by %foldershareName. They do not provide access to other site content, or the site's source code, configuration files, or other files at the site.",
    [
      '%foldershareName' => $foldershareName,
    ]) . '</p>';
  $help .= '<p>' . t(
    'Users must log into the site via a REST client. This requires providing appropriate login credentials. The server performs the same authentication procedures used by browser access.') . '</p>';

  //
  // Permissions
  // -----------
  // Explain role-based permissions.
  $help .= '<h4>' . t('Permissions') . '</h4>';
  $help .= '<p>' . t(
    'This module does not add any new permissions. Users with permission to use the %foldersharePermissionsLink module have permission to use REST web services. Access to files and folders is also governed by shared view and author grants set up in the %foldershareName web user interface.',
    [
      '%foldersharePermissionsLink' => $foldersharePermissionsLink,
      '%foldershareName' => $foldershareName,
    ]) . '</p>';

  //
  // REST configuration
  // ------------------
  // Default and recommended configuration.
  $help .= '<h4>' . t('REST configuration') . '</h4>';
  $help .= '<p>' . t(
    "The Drupal core %restHelpLink module is required by this module's REST plugin. A default configuration is included. This configuration may be modified using the contributed %restUiLink module.",
    [
      '%restHelpLink' => $restHelpLink,
      '%restUiLink'   => $restUiLink,
    ]) . '</p>';

  $help .= '<p>' . t(
    'The default and recommended configuration enables:') . '</p>';
  $help .= '<ul>';
  $help .= '<li>' . t('All HTTP methods, including GET, DELETE, PATCH, and POST.') . '</li>';
  $help .= '<li>' . t('JSON serialization.') . '</li>';
  $help .= '<li>' . t('Cookie-based authentication.') . '</li>';
  $help .= '</ul>';
  $help .= '<p>' . t(
    'The HTTP GET method is required, while the rest are optional but recommended. If DELETE is disabled, files and folders cannot be deleted. If POST is disabled, files cannot be uploaded. And if PATCH is disabled, files and folders cannot be modified.') . '</p>';

  //
  // REST client
  // -----------
  // Default and recommended configuration.
  $help .= '<h4>' . t('REST PHP client') . '</h4>';
  $help .= '<p>' . t(
    'The "client" subdirectory in this module contains a PHP command-line client named "foldershare" that uses REST connections to access %foldershareName files. This script may be distributed to a site\'s users.',
    [
      '%foldershareName' => $foldershareName,
    ]) . '</p>';

  $help .= '<h5>' . t('REST PHP client requirements') . '</h5>';
  $help .= '<p>' . t(
    'The client requires:') . '</p>';
  $help .= '<ul>';
  $help .= '<li>' . t(
    "A recent version of PHP available on the user's host.") . '</li>';
  $help .= '<li>' . t(
    "A site REST configuration with JSON or JSON+HAL serialization enabled. XML serialization is not supported.") . '</li>';
  $help .= '<li>' . t(
    "A site with standard cookie-based authentication. Drupal core's <em>Basic Authentication</em> is supported, but not recommended because it is much slower and core's implementation has limitations that make it unreliable.") . '</li>';
  $help .= '</ul>';

  $help .= '<h5>' . t('REST PHP client use') . '</h5>';
  $help .= '<p>' . t(
    'From a terminal window, help is available:') . '</p>';
  $help .= '<pre>./foldershare --help</pre>';

  $help .= '<p>' . t(
    'To connect to a site, the site\'s name must be given using the "--host" option. Login credentials may be given using "--username" and "--password" options. The rest of the command line provides a REST command to execute, such as "ls".') . '</p>';
  $help .= '<pre>./foldershare --host MYHOST --username MYNAME --password MYPASSWORD ls</pre>';

  $help .= '<p>' . t(
    'If a user name is provided, but no password, the user will be prompted for the password. If a user name is not provided, the user will access the site anonymously, just as they would by using a web browser without logging into the site.') . '</p>';

  $help .= '<p>' . t(
    'If no REST command is given, the script will enter an interactive mode that prompts for commands.') . '</p>';
  $help .= "<pre>./foldershare --host MYHOST</pre>";

  $help .= '<p>' . t(
    'Many Linux/BSD/macOS-style commands are supported, such as:') . '</p>';
  $help .= '<ul>';
  $help .= '<li>' . t("'cp' to copy a file or folder.") . '</li>';
  $help .= '<li>' . t("'hostname' to get the name of the remote host.") . '</li>';
  $help .= '<li>' . t("'ls' to list a folder's contents.") . '</li>';
  $help .= '<li>' . t("'mkdir' to create an empty folder.") . '</li>';
  $help .= '<li>' . t("'mv' to move or rename a file or folder.") . '</li>';
  $help .= '<li>' . t("'rm' to remove a file or folder.") . '</li>';
  $help .= '<li>' . t("'rmdir' to remove an empty folder.") . '</li>';
  $help .= '<li>' . t("'stat' to get file or folder attributes.") . '</li>';
  $help .= '<li>' . t("'whoami' to get the user's name.") . '</li>';
  $help .= '<li>' . t("'xattr' to get extended file attributes.") . '</li>';
  $help .= '</ul>';

  $help .= '<p>' . t(
    'Many Linux/BSD/macOS-style command options are also supported. For instance, "ls -l" gives a long folder listing and "rm -r" deletes a folder tree recursively. To get help on a command, add the "--help" option (e.g. "ls --help").') . '</p>';

  $help .= '<p>' . t(
    'Several custom commands are supported, such as:') . '</p>';
  $help .= '<ul>';
  $help .= '<li>' . t("'download' to download files and folders.") . '</li>';
  $help .= '<li>' . t("'update' to change attributes of a file or folder.") . '</li>';
  $help .= '<li>' . t("'upload' to upload files and folders.") . '</li>';
  $help .= '</ul>';

  //
  // Render element
  // --------------
  // Return a render element instead of the HTML string itself. This is
  // needed in order to include module libraries to style the help page.
  $page = [
    '#attached'  => [
      'library'  => [
        'foldershare/foldershare.module',
        'foldershare/foldershare.module.admin',
      ],
    ],
    'page'       => [
      '#type'    => 'container',
      '#tree'    => TRUE,
      '#attributes' => [
        'class'     => ['foldershare-admin-help'],
      ],
      'branding' => Branding::getBannerBranding(),
      'help'     => [
        '#type'  => 'html_tag',
        '#tag'   => 'div',
        '#value' => $help,
      ],
    ],
  ];

  return $page;
}
